from __future__ import annotations
import logging
from typing import TYPE_CHECKING, Any, Callable, Generic, Literal, Optional, Union, overload

from collections.abc import Sequence, Coroutine

import discord

from ._types import PageT
from . import utils as _utils

if TYPE_CHECKING:
    from typing_extensions import Self

    from ._types import PaginatorCheck, BaseKwargs, Destination
else:
    PaginatorCheck = Callable[[Any, discord.Interaction[Any]], Union[bool, Coroutine[Any, Any, bool]]]


__all__ = ("BaseClassPaginator",)


class BaseClassPaginator(discord.ui.View, Generic[PageT]):
    """Base class for all paginators.

    Parameters
    -----------
    pages: Sequence[Any]
        A sequence of pages to paginate.
        Supported types for pages:

        - :class:`str`: Will be set as the content of the message.
        - :class:`.discord.Embed`: Will be appended to the embeds of the message.
        - :class:`.discord.File`: Will be appended to the files of the message.
        - :class:`.discord.Attachment`: Calls :meth:`~discord.Attachment.to_file()` and appends it to the files of the message.
        - :class:`dict`: Will be updated with the kwargs of the message.
        - Sequence[Any]: Will be flattened and each entry will be handled as above.

        Sequence = List, Tuple, etc.

        Any other types will probably be ignored.
        This attribute *should* be able to be set after the paginator is created.
        Aka, hotswapping the pages.
    per_page: :class:`int`
        The amount of pages to display per page.
        Defaults to ``1``.

        E,g: If ``per_page`` is ``2`` and ``pages`` is ``["1", "2", "3", "4"]``, then the message
        will show ``["1", "2"]`` on the first page and ``["3", "4"]`` on the second page.
    author_id: Optional[:class:`int`]
        The id of the user who can interact with the paginator.
        Defaults to ``None``.
    check: Optional[Callable[[:class:`.BaseClassPaginator`, :class:`discord.Interaction`], Union[:class:`bool`, Coroutine[Any, Any, :class:`bool`]]]]
        A callable that checks if the interaction is valid. This must be a callable that takes 2 or 3 parameters.
        The last two parameters represent the interaction and paginator respectively.
        It CAN be a coroutine.

        This is called in :meth:`~discord.ui.View.interaction_check`.

        If ``author_id`` is not ``None``, this won't be called.
        Defaults to ``None``.
    always_allow_bot_owner: :class:`bool`
        Whether to always allow the bot owner to interact with the paginator.
        Defaults to ``True``.
    delete_after: :class:`bool`
        Whether to delete the message after the paginator stops. Only works if ``message`` is not ``None``.
        Defaults to ``False``.
    disable_after: :class:`bool`
        Whether to disable the paginator after the paginator stops. Only works if ``message`` is not ``None``.
        Defaults to ``False``.
    clear_buttons_after: :class:`bool`
        Whether to clear the buttons after the paginator stops. Only works if ``message`` is not ``None``.
        Defaults to ``False``.
    message: Optional[:class:`discord.Message`]
        The message to use for the paginator. This is set automatically when ``_send`` is called.
        Defaults to ``None``.
    add_page_string: :class:`bool`
        Whether to add the page string to the page. Defaults to ``True``.
        This is a string that represents the current page and the max pages. E,g: ``"Page 1 of 2"``.

        If the page is an embed, it will be appended to the footer text.
        If the page is a string, it will be appended to the string.
        else, it will be set as the content of the message.

    timeout: Optional[Union[:class:`int`, :class:`float`]]
        The timeout for the paginator.
        Defaults to ``180.0``.
    """

    def __init__(
        self,
        pages: Sequence[PageT],
        *,
        per_page: int = 1,
        author_id: Optional[int] = None,
        check: Optional[PaginatorCheck[Self]] = None,
        always_allow_bot_owner: bool = True,
        delete_after: bool = False,
        disable_after: bool = False,
        clear_buttons_after: bool = False,
        message: Optional[discord.Message] = None,
        add_page_string: bool = True,
        timeout: Optional[Union[int, float]] = 180.0,
    ) -> None:
        super().__init__(timeout=timeout)

        if not pages:
            raise ValueError("No pages provided.")
        if per_page < 1:
            raise ValueError("per_page must be greater than 0.")

        self.pages: Sequence[PageT] = pages
        self.per_page: int = per_page
        self.max_pages: int = len(pages) // per_page + bool(len(pages) % per_page)

        self._current_page: int = 0

        self.author_id: Optional[int] = author_id
        self._check: Optional[PaginatorCheck[Self]] = None
        if check is not None:
            if not callable(check) or not _utils._check_parameters_amount(check, (2, 3)):
                raise TypeError(
                    (
                        "check must be a callable with exactly 2 or 3 parameters. Last two "
                        "representing the interaction and paginator. `(async) def check(self, interaction, "
                        "paginator):` or `(async) def check(interaction, paginator):`."
                    )
                )

        self.always_allow_bot_owner: bool = always_allow_bot_owner
        self.delete_after: bool = delete_after
        self.disable_after: bool = disable_after
        self.clear_buttons_after: bool = clear_buttons_after
        self.add_page_string: bool = add_page_string

        self.message: Optional[discord.Message] = message

        self.__owner_ids: Optional[set[int]] = None

        self._reset_base_kwargs()

    async def __is_bot_owner(self, interaction: discord.Interaction[Any]) -> bool:
        """Checks if the interaction's user is one of the bot owners."""
        if self.__owner_ids is None:
            self.__owner_ids = await _utils._fetch_bot_owner_ids(interaction.client)

        return interaction.user.id in self.__owner_ids

    def _reset_base_kwargs(self) -> None:
        """Resets the base kwargs.

        This sets the base kwargs to ``{"content": None, "embeds": [], "view": self}``.
        """
        self.__base_kwargs: BaseKwargs = {"content": None, "embeds": [], "view": self}

    @property
    def current_page(self) -> int:
        """:class:`int`: The current page. Starts from ``0``."""
        return self._current_page

    @current_page.setter
    def current_page(self, value: int) -> None:
        """:class:`int`: Sets the current page to the given value."""
        self._current_page = max(0, min(value, self.max_pages - 1))

    @property
    def page_string(self) -> str:
        """:class:`str`: A string representing the current page and the max pages."""
        return f"Page {self.current_page + 1} of {self.max_pages}"

    def stop(self) -> None:
        """Stops the view and resets the base kwargs."""
        self._reset_base_kwargs()
        self.message = None
        return super().stop()

    async def on_timeout(self) -> None:
        """This method is called when the paginator times out.

        This method does the following checks (in order):
        - Calls :meth:`.BaseClassPaginator.stop_paginator`.
        - Calls :meth:`discord.ui.View.on_timeout`.
        """
        await self.stop_paginator()
        await super().on_timeout()

    async def _handle_checks(self, interaction: discord.Interaction[Any]) -> bool:
        """Handles the checks for the paginator.

        This is called in :meth:`~discord.ui.View.interaction_check`.

        Parameters
        ----------
        interaction: :class:`discord.Interaction`
            The interaction to check.

        Returns
        -------
        :class:`bool`
            Whether the interaction is valid or not.
        """
        if self.always_allow_bot_owner:
            if await self.__is_bot_owner(interaction):
                return True

        if self.author_id is not None and interaction.user.id == self.author_id:
            return True

        if self._check is not None:
            return await discord.utils.maybe_coroutine(self._check, self, interaction)

        return await super().interaction_check(interaction)

    async def interaction_check(self, interaction: discord.Interaction[Any]) -> bool:
        """This method is called by the library when the paginator receives an interaction.

        This method does the following checks (in order):

        - If ``always_allow_bot_owner`` is ``True``, it checks if the interaction's author id is one of the bot owners.
        - If ``author_id`` is not ``None``, it checks if the interaction's author id is the same as the one set.
        - If ``check`` is not ``None``, it calls it and checks if it returns ``True``.
        - If none of the above checks are ``True``, it calls :meth:`discord.ui.View.interaction_check`.

        Parameters
        ----------
        interaction: :class:`discord.Interaction`
            The interaction received.
        """
        return await self._handle_checks(interaction)

    def _do_format_page(self, page: Union[PageT, Sequence[PageT]]) -> Coroutine[Any, Any, Union[PageT, Sequence[PageT]]]:
        return discord.utils.maybe_coroutine(self.format_page, page)

    async def format_page(self, page: Union[PageT, Sequence[PageT]]) -> Union[PageT, Sequence[PageT]]:
        """This method can be overridden to format the page before sending it.
        By default, it returns the page as is.

        Parameters
        ----------
        page: Union[Any], Sequence[Any]]
            The page to format.

        Returns
        -------
        Union[Any], Sequence[Any]]
            The formatted page(s).
        """
        return page

    async def stop_paginator(self, interaction: Optional[discord.Interaction[Any]] = None) -> None:
        """Stops the paginator.

        This method does handles deleting the message, disabling the paginator and clearing the buttons.

        Parameters
        ----------
        interaction: Optional[:class:`discord.Interaction`]
            Optionally, the last interaction to edit. If ``None``, ``.message`` is used.
        """
        success: bool = False
        if self.delete_after:
            error_message: str = "Failed to delete the message. in {cls.__name__}.stop_paginator.\nError: {message}"
            to_call = (
                [
                    self.message.delete,
                ]
                if self.message
                else []
            )
            if interaction:
                to_call = [
                    interaction.delete_original_response,
                ] + to_call
                if not interaction.response.is_done():
                    await interaction.response.defer()
                if interaction.message:
                    to_call.insert(1, interaction.message.delete)

            success, error = await _utils._call_and_ignore(
                to_call,
            )
            if not success:
                logging.debug(error_message.format(cls=self.__class__, message=error))

            self.stop()
            return

        if self.disable_after or self.clear_buttons_after:
            if self.clear_buttons_after:
                self.clear_items()
            else:
                self._disable_all_children()

            to_call: list[Callable[..., Any]] = (
                [
                    self.message.edit,
                ]
                if self.message
                else []
            )
            error_message = "Failed to edit the message. in {cls.__name__}.stop_paginator.\nError: {message}"
            if interaction:
                to_call = [
                    interaction.edit_original_response,
                ] + to_call
                if not interaction.response.is_done():
                    to_call.insert(0, interaction.response.edit_message)
                if interaction.message:
                    to_call.insert(1, interaction.message.edit)

            success, error = await _utils._call_and_ignore(to_call, view=self)
            if not success:
                logging.debug(error_message.format(cls=self.__class__, message=error))

            self.stop()

    def get_page(self, page_number: int) -> Union[PageT, Sequence[PageT]]:
        """Gets the page with the given page number.

        Parameters
        ----------
        page_number: :class:`int`
            The page number to get.

        Returns
        -------
        Union[Any], Sequence[Any]]
            The page(s) with the given page number.
        """
        # handle per_page
        if page_number < 0 or page_number >= self.max_pages:
            self.current_page = 0
            return self.pages[self.current_page]

        if self.per_page == 1:
            return self.pages[page_number]
        else:
            base = page_number * self.per_page
            return self.pages[base : base + self.per_page]

    def _handle_page_string(self) -> None:
        if not self.add_page_string:
            return

        embeds = self.__base_kwargs.get("embeds", [])
        content = self.__base_kwargs.get("content")
        if embeds:
            for embed in embeds:
                to_set = self.page_string
                if footer_text := embed.footer.text:
                    if "|" in footer_text:
                        footer_text = footer_text.split("|")[0].strip()
                        to_set = f"{footer_text} | {self.page_string}"

                embed.set_footer(text=to_set)
        elif content:
            self.__base_kwargs["content"] = f"{content}\n{self.page_string}"
        else:
            self.__base_kwargs["content"] = self.page_string

    async def get_page_kwargs(self, page: Union[PageT, Sequence[PageT]], /, skip_formatting: bool = False) -> BaseKwargs:
        """Gets the kwargs to send the page with.

        Parameters
        ----------
        page: Union[Any, Sequence[Any]]
            The page to get the kwargs for.
        skip_formatting: bool
            Whether to not call :meth:`.BaseClassPaginator.format_page` with the given page.
            Defaults to ``False``.

        Returns
        -------
        :class:`.BaseKwargs`
            The kwargs to send the page with.
        """
        if not skip_formatting:
            self._reset_base_kwargs()
            _page = await self._do_format_page(page)
            return await self.get_page_kwargs(_page, skip_formatting=True)

        # Sequence
        if isinstance(page, (list, tuple)):
            inner_page: Any
            for inner_page in page:
                # handles the page kwargs
                await self.get_page_kwargs(inner_page, skip_formatting=True)

        if isinstance(page, (int, str)):
            if self.__base_kwargs["content"]:
                self.__base_kwargs["content"] += str(page)
            else:
                self.__base_kwargs["content"] = str(page)
        elif isinstance(page, discord.Embed):
            self.__base_kwargs["embeds"].append(page)
        elif isinstance(page, (discord.File, discord.Attachment)):
            # you might be wondering, why another var for this?
            # well, because of the following type error:
            # "File" is incompatible with "Sequence[PageT@BaseClassPaginator]"
            if isinstance(page, discord.Attachment):
                file = await page.to_file()
            else:
                file = page

            if "files" not in self.__base_kwargs:
                self.__base_kwargs["files"] = [file]
            else:
                self.__base_kwargs["files"].append(file)
        elif isinstance(page, dict):
            # kinda the same thing as above but it didn't appricate that it
            # didn't know the type of the key&value so it was "dict[Unknown, Unknown]"
            data: dict[Any, Any] = page.copy()
            self.__base_kwargs.update(data)

        return self.__base_kwargs

    def _disable_all_children(self) -> None:
        for child in self.children:
            if hasattr(child, "disabled"):
                child.disabled = True  # type: ignore # not all children have disabled attr.

    async def _edit_message(self, interaction: Optional[discord.Interaction[Any]] = None, /, **kwargs: Any) -> None:
        """Edits the message with the given kwargs.

        Parameters
        ----------
        interaction: Optional[:class:`discord.Interaction`]
            The interaction to edit.
            If ``None``, ``.message`` is used, if that's ``None``, a :exc:`ValueError` is raised.
            Defaults to ``None``.
        **kwargs: Any
            The kwargs to edit the message with.

        Raises
        ------
        ValueError
            If ``interaction`` is ``None`` and :attr:`.BaseClassPaginator.message` is ``None``.
        """
        kwargs.pop("ephemeral", None)
        kwargs["attachments"] = kwargs.pop("files", [])

        if interaction is None:
            if self.message is None:
                raise ValueError("interaction and self.message are both None.")
            await self.message.edit(**kwargs)

        else:
            if interaction.response.is_done():
                try:
                    await interaction.edit_original_response(**kwargs)
                except Exception:
                    try:
                        await interaction.message.edit(**kwargs)  # type: ignore
                    except Exception:
                        pass
            else:
                await interaction.response.edit_message(**kwargs)

        if self.is_finished():
            await self.stop_paginator()

    async def switch_page(self, interaction: Optional[discord.Interaction[Any]], page_number: int) -> None:
        """Switches the page to the given page number.

        Parameters
        ----------
        interaction: Optional[:class:`discord.Interaction`]
            The interaction to edit. If ``None``, ``.message`` is used.
        page_number: :class:`int`
            The page number to switch to.
        """
        self.current_page = page_number
        page = self.get_page(self.current_page)
        page_kwargs = await self.get_page_kwargs(page)
        self._handle_page_string()
        await self._edit_message(interaction, **page_kwargs)

    @overload
    async def send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: bool = ...,
        edit_message: Literal[True] = ...,
        **send_kwargs: Any,
    ) -> None:
        ...

    @overload
    async def send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: bool = ...,
        edit_message: Literal[False] = ...,
        **send_kwargs: Any,
    ) -> discord.Message:
        ...

    @overload
    async def send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: Literal[False] = ...,
        edit_message: bool = ...,
    ) -> Optional[discord.Message]:
        ...

    @overload
    async def send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: Literal[True] = ...,
        edit_message: bool = ...,
        **send_kwargs: Any,
    ) -> Optional[discord.Message]:
        ...

    async def send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: bool = False,
        edit_message: bool = False,
        **send_kwargs: Any,
    ) -> Optional[discord.Message]:
        """Sends the message to the given destination.

        Parameters
        ----------
        destination: Union[:class:`discord.abc.Messageable`, :class:`discord.Interaction`]
            The destination to send the message to. Handles responding to the interaction if given.
        override_page_kwargs: :class:`bool`
            Whether to override the page kwargs with the given kwargs to ``send_kwargs``.
            Defaults to ``False``.
        edit_message: :class:`bool`
            Whether to edit the message instead of sending a new one.
            Defaults to ``False``.
        **send_kwargs: Any
            The kwargs to pass to the destination's send method. Only used if ``override_page_kwargs`` is ``True``.

        Returns
        -------
        Optional[:class:`discord.Message`]
            The message or response sent.
        """
        return await self._send(
            destination, override_page_kwargs=override_page_kwargs, edit_message=edit_message, **send_kwargs
        )

    async def _send(
        self,
        destination: Destination,
        *,
        override_page_kwargs: bool = False,
        edit_message: bool = False,
        **send_kwargs: Any,
    ) -> Optional[discord.Message]:
        page = self.get_page(self.current_page)
        page_kwargs: dict[str, Any] = await self.get_page_kwargs(page)  # type: ignore # TypedDict don't go well with overloads
        self._handle_page_string()
        if override_page_kwargs:
            page_kwargs |= send_kwargs

        if edit_message:
            return await self._edit_message(
                destination if isinstance(destination, discord.Interaction) else None, **page_kwargs
            )

        elif isinstance(destination, discord.Interaction):
            if destination.response.is_done():
                self.message = await destination.followup.send(**page_kwargs, wait=True)
            else:
                self.message = await destination.response.send_message(**page_kwargs)
                if not self.message:
                    self.message = await destination.original_response()

        else:
            self.message = await destination.send(**page_kwargs)

        return self.message



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    Optional,
    Union,
)
from collections.abc import Sequence

from copy import deepcopy
import os

import discord

from .base_paginator import BaseClassPaginator
from ._types import PageT

if TYPE_CHECKING:
    from typing_extensions import Self, Unpack

    from ._types import BasePaginatorKwargs


__all__ = (
    "PaginatorOption",
    "SelectOptionsPaginator",
)


class PaginatorOption(discord.SelectOption, Generic[PageT]):
    """A subclass of :class:`discord.SelectOption` representing a page in a :class:`SelectOptionsPaginator`.

    Other parameters are the same as :class:`~.discord.SelectOption`.

    Parameters
    ----------
    content: Union[Any, Sequence[Any]]
        The content of the page. See :class:`SelectOptionsPaginator` for more the supported types.
    emoji: Optional[Union[:class:`str`, :class:`discord.Emoji`, :class:`discord.PartialEmoji`]]
        The emoji to use for the option. Defaults to ``None``.
    """

    def __init__(
        self,
        content: Union[PageT, Sequence[PageT]],
        *,
        label: str = discord.utils.MISSING,
        emoji: Optional[Union[str, discord.Emoji, discord.PartialEmoji]] = None,
        value: str = discord.utils.MISSING,
        description: Optional[str] = None,
    ) -> None:
        super().__init__(label=label, value=value, description=description, emoji=emoji)
        self.content: Union[PageT, Sequence[PageT]] = content

    @classmethod
    def _from_page(
        cls,
        page: Any,
        copy_from: Optional[discord.SelectOption] = None,
        *,
        label: str = discord.utils.MISSING,
        value: str = discord.utils.MISSING,
        description: Optional[str] = None,
        emoji: Optional[Union[str, discord.Emoji, discord.PartialEmoji]] = None,
    ) -> Self:
        if isinstance(page, discord.SelectOption):
            raise TypeError(
                "A regular SelectOption is not supported as it doesn't contain any content. Use PaginatorOption instead or pass a Page."
            )

        if not copy_from and label is discord.utils.MISSING and emoji is None:
            raise TypeError("Either label or emoji must be provided if copy_from is not provided")

        return cls(
            content=page,
            label=label or copy_from.label,  # type: ignore # it's handled in the if statement above
            value=value or copy_from.value,  # type: ignore # it's handled in the if statement above
            description=description or copy_from.description,  # type: ignore # it's handled in the if statement above
            emoji=emoji or copy_from.emoji,  # type: ignore # it's handled in the if statement above
        )


class SelectOptionsPaginator(BaseClassPaginator[PageT]):
    """A paginator that uses discord.ui.Select to select pages.

    This paginator provides one select and two buttons to navigate through "pages" with options.

    Just for clarification, here is a list of supported pages:

    - :class:`discord.Embed`
    - :class:`discord.File`
    - :class:`discord.Attachment`
    - :class:`str`
    - :class:`list` or :class:`tuple` of the above
    - :class:`PaginatorOption` (a subclass of :class:`discord.SelectOption` with a ``content`` kwarg that can be any of the above)

    Other parameters are the same as :class:`discord.ext.paginators.base_paginator.BaseClassPaginator`. Except ``per_page`` which is always ``1`` and cannot be changed.

    Parameters
    ----------
    pages: Sequence[Union[Any, :class:`.PaginatorOption`]]
        A sequence of pages to paginate through. A page can be anything that is supported by the :class:`.BaseClassPaginator` class.
        With the addition of :class:`PaginatorOption` which is a subclass of :class:`discord.SelectOption` that also contains a ``content`` attribute,
        it can be used to provide a custom label, description and emoji for each "page".

        All options will be split into chunks of ``per_select`` and each chunk will be a select.
        IF a chunk is a already a list, it will be treated as a single select. If this chunk contains less or more items than ``per_select``,
        it will raise a ValueError.

        Example:

        .. code-block:: python3

            pages = [
                # This will be a single select with 3 options
                # per_select must be >= 3 or it will raise a ValueError
                ["Page 1", "Page 2", "Page 3"],
                # These will span across the needed amount of selects
                "Page 4", "Page 5", "Page 6", "Page 7",
                ...
            ]
    per_select: Optional[:class:`int`]
        The amount of options per select. Defaults to :attr:`SelectOptionsPaginator.MAX_SELECT_OPTIONS`.
    default_option: Optional[:class:`discord.SelectOption`]
        The option to get the metadata from if the a page is not an instance of :class:`PaginatorOption`.
        If this is ``None``, it will try to get the metadata from the page itself.
        E,g if the page is an embed, it will try to get the title or description, etc and use that as the label.
        Defaults to ``None``.
    """

    MAX_SELECT_OPTIONS: ClassVar[Literal[25]] = 25
    """The maximum amount of options per select by discord by the time of writing this."""

    def __init__(
        self,
        pages: Sequence[
            Union[
                PageT,
                PaginatorOption[PageT],
                Sequence[PageT],
                Sequence[PaginatorOption[PageT]],
            ]
        ],
        *,
        per_select: Optional[int] = None,
        default_option: Optional[discord.SelectOption] = None,
        **kwargs: Unpack[BasePaginatorKwargs],
    ) -> None:
        self.per_select: int = per_select or self.MAX_SELECT_OPTIONS
        self.default_option: Optional[discord.SelectOption] = default_option

        if "per_page" in kwargs:
            raise TypeError("per_page cannot be changed for SelectOptionsPaginator")

        kwargs["per_page"] = 1
        super().__init__(
            self._construct_options(pages),  # type: ignore
            **kwargs,
        )
        self.current_option_index: int = 0

    def _ensure_unique_value(self, option: Optional[discord.SelectOption]) -> Optional[discord.SelectOption]:
        if not option:
            return None

        if option.value != option.label:
            return option

        new_option = deepcopy(option)
        new_option.value += str(os.urandom(16).hex()) if option.value == option.label else ""
        new_option.value = new_option.value[:99]
        return new_option

    def _get_default_option_per_page(self, page: Any) -> discord.SelectOption:
        label: Optional[str] = None
        if isinstance(page, discord.Embed):
            label = page.title or page.description or page.author.name or page.footer.text
        elif isinstance(page, (discord.File, discord.Attachment)):
            label = page.filename
        elif isinstance(page, (int, str)):
            label = str(page)

        if label:
            label = str(label.split("\n")[0][:99])
        return discord.SelectOption(label=label or "Untitled")

    def _construct_options(self, pages: Sequence[Sequence[PageT]]) -> list[list[PaginatorOption[PageT]]]:
        def actual_construct(page: Union[PageT, PaginatorOption[PageT]]) -> PaginatorOption[PageT]:
            if isinstance(page, PaginatorOption):
                return page  # type: ignore # it's a PaginatorOption

            return PaginatorOption._from_page(
                page,
                self._ensure_unique_value(self.default_option or self._get_default_option_per_page(page)),
            )

        res: list[list[PaginatorOption[PageT]]] = []
        options: list[PaginatorOption[PageT]] = []
        nested_options: list[PaginatorOption[PageT]] = []

        for page in pages:
            # Sequence
            if isinstance(page, (list, tuple)):
                if len(page) == 0:
                    continue

                for item in page:
                    if isinstance(item, (list, tuple)):
                        raise TypeError("Nested lists are not supported")

                    nested_options.append(actual_construct(item))

                if len(nested_options) > self.per_select:
                    raise ValueError(
                        f"Too many options for one select in nested list (max: {self.per_select}, got: {len(nested_options)})"
                    )

                options.append(nested_options)  # type: ignore # dw, it's a list of PaginatorOption
                nested_options = []
            else:
                options.append(actual_construct(page))  # type: ignore # Sequence is handled above

        for option in options:
            if isinstance(option, list):
                res.append(option)
                options.remove(option)

        if len(options) > 0:
            res.extend(discord.utils.as_chunks(options, self.per_select))

        return res

    def get_page(self, page_number: int) -> Union[PageT, Sequence[PageT]]:
        page: Sequence[PaginatorOption[PageT]] = super().get_page(page_number)  # type: ignore # it's a list of PaginatorOption

        option: PaginatorOption[PageT] = page[self.current_option_index]
        option.default = True
        return option.content

    async def switch_page(self, interaction: Optional[discord.Interaction[Any]], page_number: int) -> None:
        self.current_option_index = 0
        self.select_page.options = self.pages[page_number]  # type: ignore
        for option in self.select_page.options:  # type: ignore
            option.default = False

        self.select_page.placeholder = f"Select a page | {self.page_string}"

        self.previous_page.disabled = self.current_page <= 0
        self.next_page.disabled = self.current_page >= self.max_pages - 1
        return await super().switch_page(interaction, page_number)

    async def switch_options(self, interaction: discord.Interaction[Any]) -> None:
        selected: str = self.select_page.values[0]
        for idx, option in enumerate(self.select_page.options):
            if option.label == selected or option.value == selected:
                self.current_option_index = idx
                option.default = True
            else:
                option.default = False

        page = self.get_page(self.current_page)
        self.select_page.options = self.pages[self.current_page]  # type: ignore
        kwrgs = await self.get_page_kwargs(page)
        self._handle_page_string()
        await self._edit_message(interaction, **kwrgs)

    @discord.ui.button(label="Previous", style=discord.ButtonStyle.blurple, row=1)
    async def previous_page(self, interaction: discord.Interaction[Any], _: discord.ui.Button[Self]) -> None:
        if self.current_page <= 0:
            self.current_page = 0
        else:
            self.current_page -= 1

        await self.switch_page(interaction, self.current_page)

    @discord.ui.button(label="Next", style=discord.ButtonStyle.blurple, row=1)
    async def next_page(self, interaction: discord.Interaction[Any], _: discord.ui.Button[Self]) -> None:
        if self.current_page >= self.max_pages - 1:
            self.current_page = self.max_pages - 1
        else:
            self.current_page += 1

        await self.switch_page(interaction, self.current_page)

    @discord.ui.select(placeholder="Select a page", row=0)
    async def select_page(self, interaction: discord.Interaction[Any], _: discord.ui.Select[Any]) -> None:
        await self.switch_options(interaction)

    async def _send(self, *args: Any, **kwargs: Any) -> Optional[discord.Message]:
        self.select_page.options = self.pages[0]  # type: ignore # nothing really possible to do here
        self.select_page.options[0].default = True  # type: ignore # nothing really possible to do here
        self.select_page.placeholder = f"Select a page | {self.page_string}"
        self.previous_page.disabled = True
        self.next_page.disabled = self.max_pages == 1
        return await super()._send(*args, **kwargs)



try:
    from discord.ext.modal_paginator import *  # type: ignore # noqa: F403 # it's fine
except ImportError:
    print(
        "discord.ext.modal_paginator not found. Install it with `python -m pip install -U discord-ext-modal-paginator` or use the `[modalpaginator]` option when install this package."
    )



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Literal, Optional, Union
from collections.abc import Sequence

from copy import deepcopy

from discord import ButtonStyle, Emoji, PartialEmoji
import discord
from discord.ui import Button, Modal, TextInput

from .base_paginator import BaseClassPaginator
from ._types import PageT

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from ._types import BasePaginatorKwargs

    ValidButtonKeys = Literal["FIRST", "LEFT", "RIGHT", "LAST", "STOP", "PAGE_INDICATOR"]
    ValidButtonsDict = dict[ValidButtonKeys, "PaginatorButton"]

    BaseClassPaginator = BaseClassPaginator

__all__: tuple[str, ...] = ("ButtonPaginator", "PaginatorButton")


class ChooseNumber(Modal):
    number_input: TextInput[Any] = TextInput(
        placeholder="Current: {0}",
        label="Enter a number between 1 and {0}",
        custom_id="paginator:textinput:choose_number",
        max_length=0,
        min_length=1,
    )

    def __init__(self, paginator: ButtonPaginator[Any], /, **kwargs: Any) -> None:
        super().__init__(
            title="Which page would you like to go to?",
            timeout=paginator.timeout,
            custom_id="paginator:modal:choose_number",
            **kwargs,
        )
        self.paginator: ButtonPaginator[Any] = paginator
        self.number_input.max_length = paginator.max_pages
        self.number_input.label = self.number_input.label.format(paginator.max_pages)

        # type checker
        if not self.number_input.placeholder:
            self.number_input.placeholder = f"Current: {paginator.current_page + 1}"
        else:
            self.number_input.placeholder = self.number_input.placeholder.format(paginator.current_page + 1)

        self.value: Optional[int] = None

    async def on_submit(self, interaction: discord.Interaction[Any]) -> None:
        # can't happen but type checker
        if not self.number_input.value:
            await interaction.response.send_message("Please enter a number!", ephemeral=True)
            self.stop()
            return

        if (
            not self.number_input.value.isdigit()
            or int(self.number_input.value) <= 0
            or int(self.number_input.value) > self.paginator.max_pages
        ):
            await interaction.response.send_message(
                f"Please enter a valid number between 1 and {self.paginator.max_pages}", ephemeral=True
            )
            self.stop()
            return

        number = int(self.number_input.value) - 1

        if number == self.paginator.current_page:
            await interaction.response.send_message("That is the current page!", ephemeral=True)
            self.stop()
            return

        self.value = number
        await interaction.response.defer()
        self.stop()


class PageSwitcherAndStopButtonView(discord.ui.View):
    STOP: Optional[Button["ButtonPaginator[Any]"]] = None  # filled in _add_buttons
    PAGE_INDICATOR: Optional[Button["ButtonPaginator[Any]"]] = None  # filled in _add_buttons

    def __init__(self, paginator: ButtonPaginator[Any], /) -> None:
        super().__init__(timeout=paginator.timeout)

    def _add_buttons(self, paginator: ButtonPaginator[Any], /) -> None:
        self._paginator: ButtonPaginator[Any] = paginator
        if not any(key in ("STOP", "PAGE_INDICATOR") for key in paginator._buttons):
            raise ValueError("STOP and PAGE_INDICATOR buttons are required if combine_switcher_and_stop_button is True.")

        org_page_indicator_button: PaginatorButton = paginator._buttons["PAGE_INDICATOR"]
        page_indicator_button = PaginatorButton(
            label="Switch Page",
            emoji=org_page_indicator_button.emoji,
            style=org_page_indicator_button.style,
            custom_id="switch_page",
            disabled=False,
        )

        org_stop_button = paginator._buttons["STOP"]
        stop_button = PaginatorButton(
            label=org_stop_button.label,
            emoji=org_stop_button.emoji,
            style=org_stop_button.style,
            custom_id="stop_button",
            disabled=False,
        )
        buttons: dict[str, PaginatorButton] = {
            "STOP": stop_button,
            "PAGE_INDICATOR": page_indicator_button,
        }
        for name, button in buttons.items():
            setattr(self, name, button)
            self.add_item(button)

    async def callback(self, interaction: discord.Interaction[Any], button: PaginatorButton) -> None:
        if button.custom_id == "stop_button":
            await interaction.response.defer()
            await interaction.delete_original_response()
            await self._paginator.stop_paginator(None)
            return

        if button.custom_id == "switch_page":
            new_page = await self._paginator._handle_modal(interaction)
            await interaction.delete_original_response()
            if new_page is not None:
                self._paginator.current_page = new_page
            else:
                return

        await self._paginator.switch_page(None, self._paginator.current_page)


class PaginatorButton(Button[Union["ButtonPaginator[Any]", PageSwitcherAndStopButtonView]]):
    """A button for the paginator.

    This class has a few parameters that differ from the base button.
    This can can be used passed to the ``buttons`` parameter in :class:`.ButtonPaginator`
    to customize the buttons used.

    See other parameters on :class:`discord.ui.Button`.

    Parameters
    -----------
    position: Optional[:class:`int`]
        The position of the button. Defaults to ``None``.
        If not specified, the button will be placed in the order they were added
        or whatever order discord.py adds them in.
    """

    def __init__(
        self,
        *,
        emoji: Optional[Union[Emoji, PartialEmoji, str]] = None,
        label: Optional[str] = None,
        custom_id: Optional[str] = None,
        style: ButtonStyle = ButtonStyle.blurple,
        row: Optional[int] = None,
        disabled: bool = False,
        position: Optional[int] = None,
    ) -> None:
        self.__original_kwargs: dict[str, Any] = {
            "emoji": emoji,
            "label": label,
            "custom_id": custom_id,
            "style": style,
            "row": row,
            "disabled": disabled,
            "position": position,
        }
        super().__init__(emoji=emoji, label=label, custom_id=custom_id, style=style, row=row, disabled=disabled)
        self.position: Optional[int] = position

    async def callback(self, interaction: discord.Interaction[Any]) -> None:
        # type checker
        if not self.view:
            raise ValueError("Something went wrong... button.view is None")

        if isinstance(self.view, PageSwitcherAndStopButtonView):
            await self.view.callback(interaction, self)
            return

        if self.custom_id == "stop_button":
            await self.view.stop_paginator(interaction)
            return

        if self.custom_id == "right_button":
            self.view.current_page += 1
        elif self.custom_id == "left_button":
            self.view.current_page -= 1
        elif self.custom_id == "first_button":
            self.view.current_page = 0
        elif self.custom_id == "last_button":
            self.view.current_page = self.view.max_pages - 1
        elif self.custom_id == "page_indicator_button":
            if self.view._stop_button_and_page_switcher_view:
                await interaction.response.send_message(view=self.view._stop_button_and_page_switcher_view, ephemeral=True)
                return

            new_page = await self.view._handle_modal(interaction)
            if new_page is not None:
                self.view.current_page = new_page
            else:
                return

        await self.view.switch_page(interaction, self.view.current_page)

    def _copy(self) -> PaginatorButton:
        """Create a copy of the button.
        
        Returns
        -------
        :class:`.PaginatorButton`
            A copy of the button.
        """
        return PaginatorButton(**self.__original_kwargs)
        


class ButtonPaginator(BaseClassPaginator[PageT]):
    """A paginator that uses buttons to switch pages.

    This class has a few parameters that differ from the base paginator.

    Just for clarification, here is a list of supported pages:

    - :class:`discord.Embed`
    - :class:`discord.File`
    - :class:`discord.Attachment`
    - :class:`str`
    - :class:`dict`
    - :class:`list` or :class:`tuple` of the above

    See other parameters on :class:`.BaseClassPaginator`.

    Parameters
    ----------
    buttons: Dict[:class:`str`, :class:`.PaginatorButton`]
        A dictionary of buttons to use. The keys must be one of the following:
        "FIRST", "LEFT", "RIGHT", "LAST", "STOP", "PAGE_INDICATOR".
        The values must be a PaginatorButton or ``None`` to remove the button.
        If not specified, the default buttons will be used.

        Example
        -------
        .. code-block:: python3
            :linenos:

            from discord.ext.paginators.button_paginator import ButtonPaginator, PaginatorButton

            custom_buttons = {
                # change the label of the first button from "First" to "Go to first page"
                "FIRST": PaginatorButton(label="Go to first page"),
                # change the style of the LAST button to red
                "LAST": PaginatorButton(style=ButtonStyle.red),
            }

            # pass the custom buttons to the paginator
            paginator = ButtonPaginator(pages, buttons=custom_buttons)
            ... # rest of code

    always_show_stop_button: bool
        Whether to always show the stop button, even if there is only one page.
        Defaults to ``False``.

        .. note::
            If ``always_show_stop_button`` is ``True``, the ``STOP`` key in ``buttons`` cannot be ``None``.
    combine_switcher_and_stop_button: :class:`bool`
        Whether to combine the page switcher and stop button into the paginator indicator which will send another set
        of buttons to switch pages and stop the paginator as an ephemeral message when clicked.
        Defaults to ``False``.

        .. note::
            If ``combine_switcher_and_stop_button`` is ``True``, the ``STOP`` and ``PAGE_INDICATOR`` keys in ``buttons`` cannot be ``None``.
    **kwargs: Unpack[:class:`.BasePaginatorKwargs`]
        See other parameters on :class:`discord.ext.paginator.base_paginator.BaseClassPaginator`.
    """

    FIRST: Optional[PaginatorButton] = None  # filled in __add_buttons
    LEFT: Optional[PaginatorButton] = None  # filled in __add_buttons
    RIGHT: Optional[PaginatorButton] = None  # filled in __add_buttons
    LAST: Optional[PaginatorButton] = None  # filled in __add_buttons
    STOP: Optional[PaginatorButton] = None  # filled in __add_buttons
    PAGE_INDICATOR: Optional[PaginatorButton] = None  # filled in __add_buttons

    def __init__(
        self,
        pages: Sequence[PageT],
        *,
        buttons: ValidButtonsDict = {},
        always_show_stop_button: bool = False,
        combine_switcher_and_stop_button: bool = False,
        **kwargs: Unpack[BasePaginatorKwargs],
    ) -> None:
        """Initialize the Paginator."""
        super().__init__(pages, **kwargs)

        DEFAULT_BUTTONS: dict[ValidButtonKeys, PaginatorButton] = {
            "FIRST": PaginatorButton(label="First", position=0),
            "LEFT": PaginatorButton(label="Left", position=1),
            "PAGE_INDICATOR": PaginatorButton(label="Page N/A / N/A", position=2, disabled=False),
            "RIGHT": PaginatorButton(label="Right", position=3),
            "LAST": PaginatorButton(label="Last", position=4),
            "STOP": PaginatorButton(label="Stop", style=ButtonStyle.danger, position=5),
        }

        self._buttons: dict[ValidButtonKeys, PaginatorButton] = DEFAULT_BUTTONS.copy()
        if buttons:
            valid_keys = ", ".join(DEFAULT_BUTTONS.keys())
            error_message = (
                f"buttons must be a dictionary of keys: {valid_keys} and PaginatorButton or None "
                "to remove the button as the value. Or don't specify the kwarg to use the default buttons."
            )
            if (
                not isinstance(buttons, dict)
                or any(k not in DEFAULT_BUTTONS for k in buttons)
                or not all(not v or isinstance(v, PaginatorButton) for v in buttons.values())
            ):
                raise TypeError(error_message)

            self._buttons.update(buttons)

        self._stop_button_and_page_switcher_view: Optional[PageSwitcherAndStopButtonView] = (
            PageSwitcherAndStopButtonView(self) if combine_switcher_and_stop_button else None
        )

        self.always_show_stop_button: bool = always_show_stop_button
        self.__add_buttons()

    def __handle_always_show_stop_button(self) -> None:
        if not self.always_show_stop_button:
            return

        if "STOP" not in self._buttons:
            raise ValueError("STOP button is required if always_show_stop_button is True.")

        name = "STOP"
        button = self._buttons[name]
        button.custom_id = f"{name.lower()}_button"
        setattr(self, name, button)
        self.add_item(button)

    async def _handle_modal(self, interaction: discord.Interaction[Any]) -> Optional[int]:
        modal = ChooseNumber(self)
        await interaction.response.send_modal(modal)
        await modal.wait()
        return modal.value

    def __add_buttons(self) -> None:
        if not self.max_pages > 1:
            if self.always_show_stop_button:
                self.__handle_always_show_stop_button()
                return

            self.stop()
            return

        _buttons: dict[str, PaginatorButton] = {name: button._copy() for name, button in self._buttons.copy().items() if button}
        sorted_buttons = sorted(_buttons.items(), key=lambda b: b[1].position if b[1].position is not None else 0)
        for name, button in sorted_buttons:
            custom_id = f"{name.lower()}_button"
            button.custom_id = custom_id

            setattr(self, name, button)

            if button.custom_id == "page_indicator_button":
                button.label = self.page_string
                if self.max_pages <= 2:
                    button.disabled = True

            if button.custom_id in ("first_button", "last_button"):
                if self.max_pages <= 2:
                    continue

                label = button.label if button.label else ''
                if button.custom_id == "first_button":
                    button.label = f"1 {label}"
                else:
                    button.label = f"{label} {self.max_pages}"

            if self._stop_button_and_page_switcher_view and button.custom_id == "stop_button":
                continue

            self.add_item(button)

        if self._stop_button_and_page_switcher_view:
            self._stop_button_and_page_switcher_view._add_buttons(self)

        self._update_buttons_state()

    def _update_buttons_state(self) -> None:
        for button in self.children:
            # type checker
            if not isinstance(button, PaginatorButton):
                continue

            # type checker
            if not button.custom_id:
                raise ValueError("Something went wrong... button.custom_id is None")

            original_button = self._buttons.get(f"{button.custom_id.split('_')[0].upper()}")  # type: ignore

            if button.custom_id in ("page_indicator_button", "stop_button"):
                if button.custom_id == "page_indicator_button":
                    button.label = self.page_string
                continue

            if button.custom_id in ("right_button", "last_button"):
                button.disabled = self._current_page >= self.max_pages - 1
            elif button.custom_id in ("left_button", "first_button"):
                button.disabled = self._current_page <= 0

            if button.custom_id in ("first_button", "last_button"):
                if self.max_pages <= 2:
                    button.disabled = True

                if original_button:
                    label = original_button.label if original_button.label else ''
                    if button.custom_id == "first_button":
                        button.label = f"1 {label}"
                    else:
                        button.label = f"{label} {self.max_pages}"

            if not button.disabled:
                button.style = ButtonStyle.green
            else:
                button.style = original_button.style if original_button else ButtonStyle.secondary

    @property
    def current_page(self) -> int:
        return self._current_page

    @current_page.setter
    def current_page(self, value: int) -> None:
        self._current_page = value
        self._update_buttons_state()


    def _send(self, *args: Any, **kwargs: Any) -> Any:
        self._update_buttons_state()
        return super()._send(*args, **kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional, Union, Tuple

from collections.abc import Callable

import inspect

import discord

if TYPE_CHECKING:
    from discord.ext.commands import Bot


# kinda like discord's is_owner method on commands.Bot
# but then for Client too and without setting any attributes
# https://github.com/Rapptz/discord.py/blob/bd402b486cc12f0c1bf7377fd65f2fe0a8fabd73/discord/ext/commands/bot.py#L485-L535
async def _fetch_bot_owner_ids(client: Union[discord.Client, Bot]) -> set[int]:  # type: ignore # unused
    owner_ids: list[int] = []
    if owner_id_attr := getattr(client, "owner_id", None):
        owner_ids.append(owner_id_attr)
    if owner_ids_attr := getattr(client, "owner_ids", set[int]()):
        owner_ids.extend(owner_ids_attr)

    # support for team roles is added in dpy v2.4
    if discord.version_info >= (2, 4):
        team_class: Any = getattr(discord, "TeamMemberRole", None)
        if team_class:
            app: discord.AppInfo = client.application or await client.application_info()
            if app.team:
                owner_ids.extend(
                    m.id
                    for m in app.team.members
                    if m.role in (team_class.admin, team_class.developer) and hasattr(m, "role")  # type: ignore
                )
            else:
                owner_ids.append(app.owner.id)

    return set(owner_ids)


def _check_parameters_amount(func: Callable[..., Any], amounts: tuple[int, ...], /) -> bool:  # type: ignore # unused
    parameters = inspect.signature(func).parameters
    return len(parameters) in amounts


async def _call_and_ignore(  # type: ignore # unused
    functions: list[Callable[..., Any]],
    *args: Any,
    **kwargs: Any,
) -> Tuple[bool, Optional[Exception]]:
    if not functions:
        return True, None

    for function in functions[:-1]:
        try:
            await function(*args, **kwargs)
        except Exception:
            pass
        else:
            return True, None

    try:
        await functions[-1]()
    except Exception as e:
        return False, e
    
    return True, None
    



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypedDict,
    Union,
    TypeVar,
)

from collections.abc import Callable, Coroutine

import discord

if TYPE_CHECKING:
    from typing_extensions import Self, NotRequired

    from .base_paginator import BaseClassPaginator

    BaseClassPaginator = BaseClassPaginator[Any]
else:
    BaseClassPaginator = Any


PaginatorT = TypeVar("PaginatorT", bound=BaseClassPaginator)
PaginatorCheck = Callable[[PaginatorT, discord.Interaction[Any]], Union[bool, Coroutine[Any, Any, bool]]]
Destination = Union[discord.abc.Messageable, discord.Interaction[Any]]
PageT = TypeVar("PageT", covariant=True)


class BaseKwargs(TypedDict):
    content: Optional[str]
    """Optional[:class:`str`]: Content of the page."""
    embeds: list[discord.Embed]
    """List[:class:`discord.Embed`]: Embeds of the page."""
    view: Self
    """View of the page. (the paginator)"""

    files: NotRequired[list[Union[discord.File, discord.Attachment]]]
    """NotRequired[List[:class:`discord.File`]]: Files of the page. Not always available like when using `edit`."""
    attachments: NotRequired[list[discord.File]]  # used in edit over files
    """NotRequired[List[Union[:class:`discord.File`, :class:`discord.Attachment`]]]: Attachments of the page. Not always available, probably only when using `edit`."""
    wait: NotRequired[bool]  # webhook/followup
    """NotRequired[:class:`bool`]: Whether to wait for the webhook message to be sent and returned. Only used in interaction followups."""


class BasePaginatorKwargs(TypedDict):
    check: NotRequired[Optional[PaginatorCheck[Any]]]  # default: None
    author_id: NotRequired[Optional[int]]  # default: None
    delete_after: NotRequired[bool]  # default: False
    disable_after: NotRequired[bool]  # default: False
    clear_buttons_after: NotRequired[bool]  # default: False
    per_page: NotRequired[int]  # default: 1
    timeout: NotRequired[Optional[Union[int, float]]]  # default: 180.0
    message: NotRequired[discord.Message]  # default: None
